@@ -27,15 +27,19 @@ __copyright__ = "Copyright (c) 2005-2008 Sun Microsystems Inc." \
                 "Copyright (c) 2016 Igalia, S.L."
 __license__   = "LGPL"
 
-import pyatspi
+import gi
+gi.require_version("Atspi", "2.0")
+from gi.repository import Atspi
 import re
 
 from . import braille
 from . import debug
-from . import eventsynthesizer
-from . import messages
+from . import orca
 from . import orca_state
 from . import settings
+from .ax_event_synthesizer import AXEventSynthesizer
+from .ax_object import AXObject
+from .ax_utilities import AXUtilities
 
 EMBEDDED_OBJECT_CHARACTER = '\ufffc'
 
@@ -89,6 +93,13 @@ class Word:
         self.height = height
         self.chars = []
 
+    def __str__(self):
+        return "WORD: '%s' (%i-%i) %s" % \
+            (self.string.replace("\n", "\\n"),
+             self.startOffset,
+             self.endOffset,
+             self.zone.accessible)
+
     def __getattribute__(self, attr):
         if attr != "chars":
             return super().__getattribute__(attr)
@@ -99,14 +110,18 @@ class Word:
 
         try:
             text = self.zone.accessible.queryText()
-        except:
+        except Exception:
             text = None
 
         chars = []
         for i, char in enumerate(self.string):
             start = i + self.startOffset
             if text:
-                extents = text.getRangeExtents(start, start+1, pyatspi.DESKTOP_COORDS)
+                try:
+                    extents = text.getRangeExtents(start, start+1, Atspi.CoordType.SCREEN)
+                except Exception as error:
+                    tokens = ["FLAT REVIEW: Exception in getRangeExtents:", error]
+                    debug.printTokens(debug.LEVEL_INFO, tokens, True)
             chars.append(Char(self, i, start, char, *extents))
 
         return chars
@@ -143,9 +158,12 @@ class Zone:
         self.y = y
         self.width = width
         self.height = height
-        self.role = role or accessible.getRole()
+        self.role = role or AXObject.get_role(accessible)
         self._words = []
 
+    def __str__(self):
+        return "ZONE: '%s' %s" % (self._string.replace("\n", "\\n"), self.accessible)
+
     def __getattribute__(self, attr):
         """To ensure we update the content."""
 
@@ -172,14 +190,14 @@ class Zone:
     def _shouldFakeText(self):
         """Returns True if we should try to fake the text interface"""
 
-        textRoles = [pyatspi.ROLE_LABEL,
-                     pyatspi.ROLE_MENU,
-                     pyatspi.ROLE_MENU_ITEM,
-                     pyatspi.ROLE_CHECK_MENU_ITEM,
-                     pyatspi.ROLE_RADIO_MENU_ITEM,
-                     pyatspi.ROLE_PAGE_TAB,
-                     pyatspi.ROLE_PUSH_BUTTON,
-                     pyatspi.ROLE_TABLE_CELL]
+        textRoles = [Atspi.Role.LABEL,
+                     Atspi.Role.MENU,
+                     Atspi.Role.MENU_ITEM,
+                     Atspi.Role.CHECK_MENU_ITEM,
+                     Atspi.Role.RADIO_MENU_ITEM,
+                     Atspi.Role.PAGE_TAB,
+                     Atspi.Role.PUSH_BUTTON,
+                     Atspi.Role.TABLE_CELL]
 
         if self.role in textRoles:
             return True
@@ -210,22 +228,26 @@ class Zone:
     def onSameLine(self, zone):
         """Returns True if we treat this Zone and zone as being on one line."""
 
-        if pyatspi.ROLE_SCROLL_BAR in [self.role, zone.role]:
+        if Atspi.Role.SCROLL_BAR in [self.role, zone.role]:
             return self.accessible == zone.accessible
 
-        try:
-            thisParentRole = self.accessible.parent.getRole()
-            zoneParentRole = zone.accessible.parent.getRole()
-        except:
-            pass
-        else:
-            if pyatspi.ROLE_MENU_BAR in [thisParentRole, zoneParentRole]:
-                return self.accessible.parent == zone.accessible.parent
+        thisParent = AXObject.get_parent(self.accessible)
+        thisParentRole = AXObject.get_role(thisParent)
+        zoneParent = AXObject.get_parent(zone.accessible)
+        zoneParentRole = AXObject.get_role(zoneParent)
+        if Atspi.Role.MENU_BAR in [thisParentRole, zoneParentRole]:
+            return thisParent == zoneParent
 
         return self._extentsAreOnSameLine(zone)
 
     def getWordAtOffset(self, charOffset):
+        msg = f"FLAT REVIEW: Searching for word at offset {charOffset}"
+        debug.printMessage(debug.LEVEL_INFO, msg, True)
+
         for word in self.words:
+            tokens = ["FLAT REVIEW: Checking", word]
+            debug.printTokens(debug.LEVEL_INFO, tokens, True)
+
             offset = word.getRelativeOffset(charOffset)
             if offset >= 0:
                 return word, offset
@@ -259,14 +281,14 @@ class TextZone(Zone):
     def __getattribute__(self, attr):
         """To ensure we update the content."""
 
-        if not attr in ["words", "string"]:
+        if attr not in ["words", "string"]:
             return super().__getattribute__(attr)
 
         string = self._itext.getText(self.startOffset, self.endOffset)
         words = []
         for i, word in enumerate(re.finditer(self.WORDS_RE, string)):
             start, end = map(lambda x: x + self.startOffset, word.span())
-            extents = self._itext.getRangeExtents(start, end, pyatspi.DESKTOP_COORDS)
+            extents = self._itext.getRangeExtents(start, end, Atspi.CoordType.SCREEN)
             words.append(Word(self, i, start, word.group(), *extents))
 
         self._string = string
@@ -339,7 +361,7 @@ class ValueZone(Zone):
         rolename = generator.getLocalizedRoleName(self.accessible)
         value = generator.getValue(self.accessible)
         if rolename and value:
-            result = "%s %s" % (rolename, value[0])
+            result = f"{rolename} {value[0]}"
 
         return result
 
@@ -390,19 +412,19 @@ class Line:
                 # The 'isinstance(zone, TextZone)' test is a sanity check
                 # to handle problems with Java text. See Bug 435553.
                 if isinstance(zone, TextZone) and \
-                   ((zone.accessible.getRole() in \
-                         (pyatspi.ROLE_TEXT,  
-                          pyatspi.ROLE_PASSWORD_TEXT,
-                          pyatspi.ROLE_TERMINAL)) or \
+                   ((AXObject.get_role(zone.accessible) in \
+                         (Atspi.Role.TEXT,  
+                          Atspi.Role.PASSWORD_TEXT,
+                          Atspi.Role.TERMINAL)) or \
                     # [[[TODO: Eitan - HACK: 
                     # This is just to get FF3 cursor key routing support.
                     # We really should not be determining all this stuff here,
                     # it should be in the scripts. 
                     # Same applies to roles above.]]]
-                    (zone.accessible.getRole() in \
-                         (pyatspi.ROLE_PARAGRAPH,
-                          pyatspi.ROLE_HEADING,
-                          pyatspi.ROLE_LINK))):
+                    (AXObject.get_role(zone.accessible) in \
+                         (Atspi.Role.PARAGRAPH,
+                          Atspi.Role.HEADING,
+                          Atspi.Role.LINK))):
                     region = braille.ReviewText(zone.accessible,
                                                 zone.string,
                                                 zone.startOffset,
@@ -410,7 +432,7 @@ class Line:
                 else:
                     try:
                         brailleString = zone.brailleString
-                    except:
+                    except Exception:
                         brailleString = zone.string
                     region = braille.ReviewComponent(zone.accessible,
                                                      brailleString,
@@ -455,7 +477,7 @@ class Context:
     WRAP_TOP_BOTTOM = 1 << 1
     WRAP_ALL        = (WRAP_LINE | WRAP_TOP_BOTTOM)
 
-    def __init__(self, script):
+    def __init__(self, script, root=None):
         """Create a new Context for script."""
 
         self.script = script
@@ -468,20 +490,33 @@ class Context:
         self.targetCharInfo = None
         self.focusZone = None
         self.container = None
-        self.focusObj = orca_state.locusOfFocus
-        self.topLevel = script.utilities.topLevelObject(self.focusObj)
+        self.focusObj = orca.getActiveModeAndObjectOfInterest()[1] or orca_state.locusOfFocus
+        self.topLevel = None
         self.bounds = 0, 0, 0, 0
 
+        frame, dialog = script.utilities.frameAndDialog(self.focusObj)
+        if root is not None:
+            self.topLevel = root
+            tokens = ["FLAT REVIEW: Restricting flat review to", root]
+            debug.printTokens(debug.LEVEL_INFO, tokens, True)
+        else:
+            self.topLevel = dialog or frame
+        tokens = ["FLAT REVIEW: Frame:", frame, "Dialog:", dialog, ". Top level:", self.topLevel]
+        debug.printTokens(debug.LEVEL_INFO, tokens, True)
+
         try:
             component = self.topLevel.queryComponent()
-            self.bounds = component.getExtents(pyatspi.DESKTOP_COORDS)
-        except:
-            msg = "ERROR: Exception getting extents of %s" % self.topLevel
-            debug.println(debug.LEVEL_INFO, msg, True)
+            self.bounds = component.getExtents(Atspi.CoordType.SCREEN)
+        except Exception:
+            tokens = ["ERROR: Exception getting extents of", self.topLevel]
+            debug.printTokens(debug.LEVEL_INFO, tokens, True)
 
-        containerRoles = [pyatspi.ROLE_MENU]
-        isContainer = lambda x: x and x.getRole() in containerRoles
-        container = pyatspi.findAncestor(self.focusObj, isContainer)
+        containerRoles = [Atspi.Role.MENU]
+
+        def isContainer(x):
+            return AXObject.get_role(x) in containerRoles
+
+        container = AXObject.find_ancestor(self.focusObj, isContainer)
         if not container and isContainer(self.focusObj):
             container = self.focusObj
 
@@ -502,9 +537,11 @@ class Context:
                     self.charIndex = offset
                 break
 
-        msg = "FLAT REVIEW: On line %i, zone %i, word %i, char %i" % \
-              (self.lineIndex, self.zoneIndex, self.wordIndex, self.charIndex)
-        debug.println(debug.LEVEL_INFO, msg, True)
+        msg = (
+            f"FLAT REVIEW: On line {self.lineIndex}, zone {self.zoneIndex} "
+            f"word {self.wordIndex}, char {self.charIndex}"
+        )
+        debug.printMessage(debug.LEVEL_INFO, msg, True)
 
     def splitTextIntoZones(self, accessible, string, startOffset, cliprect):
         """Traverses the string, splitting it up into separate zones if the
@@ -518,80 +555,37 @@ class Context:
         - startOffset: the starting character offset of the string
         - cliprect: the extents that the Zones must fit inside.
 
-        Returns a list of Zones for the visible text or None if nothing is
-        visible.
+        Returns a list of Zones for the visible text.
         """
 
-        # We convert the string to unicode and walk through it.  While doing
-        # this, we keep two sets of offsets:
-        #
-        # substring{Start,End}Offset: where in the accessible text 
-        # implementation we are
-        #
-        # unicodeStartOffset: where we are in the unicodeString
-        #
-        anyVisible = False
         zones = []
-        text = accessible.queryText()
-        substringStartOffset = startOffset
-        substringEndOffset   = startOffset
-        unicodeStartOffset   = 0
-        unicodeString = string
-        #print "LOOKING AT '%s'" % unicodeString
-        for i in range(0, len(unicodeString) + 1):
-            if (i != len(unicodeString)) \
-               and (unicodeString[i] != EMBEDDED_OBJECT_CHARACTER):
-                substringEndOffset += 1
-            elif (substringEndOffset == substringStartOffset):
-                substringStartOffset += 1
-                substringEndOffset   = substringStartOffset
-                unicodeStartOffset   = i + 1
-            else:
-                extents = text.getRangeExtents(
-                    substringStartOffset, substringEndOffset, 0)
-                if self.script.utilities.containsRegion(extents, cliprect):
-                    anyVisible = True
-                    clipping = self.script.utilities.intersection(extents, cliprect)
+        substrings = [(*m.span(), m.group(0))  for m in re.finditer(r"[^\ufffc]+", string)]
+        substrings = list(map(lambda x: (x[0] + startOffset, x[1] + startOffset, x[2]), substrings))
+        for (start, end, substring) in substrings:
+            extents = accessible.queryText().getRangeExtents(start, end, Atspi.CoordType.SCREEN)
+            if self.script.utilities.containsRegion(extents, cliprect):
+                clipping = self.script.utilities.intersection(extents, cliprect)
+                zones.append(TextZone(accessible, start, substring, *clipping))
 
+        return zones
 
-                    # [[[TODO: WDW - HACK it would be nice to clip the
-                    # the text by what is really showing on the screen,
-                    # but this seems to hang Orca and the client. Logged
-                    # as bugzilla bug 319770.]]]
-                    #
-                    #ranges = text.getBoundedRanges(\
-                    #    clipping[0],
-                    #    clipping[1],
-                    #    clipping[2],
-                    #    clipping[3],
-                    #    0,
-                    #    pyatspi.TEXT_CLIP_BOTH,
-                    #    pyatspi.TEXT_CLIP_BOTH)
-                    #
-                    #print
-                    #print "HERE!"
-                    #for range in ranges:
-                    #    print range.startOffset
-                    #    print range.endOffset
-                    #    print range.content
+    def _getLines(self, accessible, startOffset, endOffset):
+        # TODO - JD: Move this into the script utilities so we can better handle
+        # app and toolkit quirks and also reuse this (e.g. for SayAll).
+        try:
+            text = accessible.queryText()
+        except NotImplementedError:
+            return []
 
-                    substring = unicodeString[unicodeStartOffset:i]
-                    #print " SUBSTRING '%s'" % substring
-                    zones.append(TextZone(accessible,
-                                          substringStartOffset,
-                                          substring,
-                                          clipping[0],
-                                          clipping[1],
-                                          clipping[2],
-                                          clipping[3]))
-                    substringStartOffset = substringEndOffset + 1
-                    substringEndOffset   = substringStartOffset
-                    unicodeStartOffset   = i + 1
+        lines = []
+        offset = startOffset
+        while offset < min(endOffset, text.characterCount):
+            result = text.getTextAtOffset(offset, Atspi.TextBoundaryType.LINE_START)
+            if result[0] and result not in lines:
+                lines.append(result)
+            offset = max(result[2], offset + 1)
 
-        if anyVisible:
-            return zones
-        else:
-            return None
+        return lines
 
     def getZonesFromText(self, accessible, cliprect):
         """Gets a list of Zones from an object that implements the
@@ -604,32 +598,22 @@ class Context:
         Returns a list of Zones.
         """
 
-        try:
-            text = accessible.queryText()
-        except NotImplementedError:
+        if not self.script.utilities.hasPresentableText(accessible):
             return []
-        else:
-            zones = []
+
+        zones = []
+        text = accessible.queryText()
 
         # TODO - JD: This is here temporarily whilst I sort out the rest
         # of the text-related mess.
-        if not re.search("[^\ufffc]", text.getText(0, -1)):
-            return []
-
-        # TODO - JD: Ditto.
-        if "EditableText" in pyatspi.listInterfaces(accessible) \
-           and accessible.getState().contains(pyatspi.STATE_SINGLE_LINE):
+        if AXObject.supports_editable_text(accessible) \
+           and AXUtilities.is_single_line(accessible):
             extents = accessible.queryComponent().getExtents(0)
             return [TextZone(accessible, 0, text.getText(0, -1), *extents)]
 
-        debug.println(debug.LEVEL_FINEST, "  looking at text:")
-
-        offset = 0
-        lastEndOffset = -1
         upperMax = lowerMax = text.characterCount
         upperMid = lowerMid = int(upperMax / 2)
         upperMin = lowerMin = 0
-        upperY = lowerY = 0
         oldMid = 0
 
         # performing binary search to locate first line inside clipped area
@@ -638,7 +622,6 @@ class Context:
             [x, y, width, height] = text.getRangeExtents(upperMid,
                                                          upperMid+1,
                                                          0)
-            upperY = y
             if y > cliprect.y:
                 upperMax = upperMid
             else:
@@ -653,79 +636,21 @@ class Context:
             [x, y, width, height] = text.getRangeExtents(lowerMid,
                                                          lowerMid+1,
                                                          0)
-            lowerY = y
             if y > limit:
                 lowerMax = lowerMid
             else:
                 lowerMin = lowerMid
             lowerMid = int((lowerMax - lowerMin) / 2) + lowerMin
 
-        # finding out the zones
-        offset = upperMin
-        length = lowerMax
-        while offset < length:
+        msg = "FLAT REVIEW: Getting lines for %s offsets %i-%i" % (accessible, upperMin, lowerMax)
+        debug.printMessage(debug.LEVEL_INFO, msg, True)
 
-            [string, startOffset, endOffset] = text.getTextAtOffset(
-                offset,
-                pyatspi.TEXT_BOUNDARY_LINE_START)
+        lines = self._getLines(accessible, upperMin, lowerMax)
+        tokens = ["FLAT REVIEW:", len(lines), "lines found for", accessible]
+        debug.printTokens(debug.LEVEL_INFO, tokens, True)
 
-            debug.println(debug.LEVEL_FINEST,
-                          "    line at %d is (start=%d end=%d): '%s'" \
-                          % (offset, startOffset, endOffset, string))
-
-            # [[[WDW - HACK: well...gnome-terminal sometimes wants to
-            # give us outrageous values back from getTextAtOffset
-            # (see http://bugzilla.gnome.org/show_bug.cgi?id=343133),
-            # so we try to handle it.  Evolution does similar things.]]]
-            #
-            if (startOffset < 0) \
-               or (endOffset < 0) \
-               or (startOffset > offset) \
-               or (endOffset < offset) \
-               or (startOffset > endOffset) \
-               or (abs(endOffset - startOffset) > 666e3):
-                debug.println(debug.LEVEL_WARNING,
-                              "flat_review:getZonesFromText detected "\
-                              "garbage from getTextAtOffset for accessible "\
-                              "name='%s' role'='%s': offset used=%d, "\
-                              "start/end offset returned=(%d,%d), string='%s'"\
-                              % (accessible.name, accessible.getRoleName(),
-                                 offset, startOffset, endOffset, string))
-                break
-
-            # [[[WDW - HACK: this is here because getTextAtOffset
-            # tends not to be implemented consistently across toolkits.
-            # Sometimes it behaves properly (i.e., giving us an endOffset
-            # that is the beginning of the next line), sometimes it
-            # doesn't (e.g., giving us an endOffset that is the end of
-            # the current line).  So...we hack.  The whole 'max' deal
-            # is to account for lines that might be a brazillion lines
-            # long.]]]
-            #
-            if endOffset == lastEndOffset:
-                offset = max(offset + 1, lastEndOffset + 1)
-                lastEndOffset = endOffset
-                continue
-            else:
-                offset = endOffset
-                lastEndOffset = endOffset
-
-            textZones = self.splitTextIntoZones(
-                accessible, string, startOffset, cliprect)
-
-            if textZones:
-                zones.extend(textZones)
-            elif len(zones):
-                # We'll break out of searching all the text - the idea
-                # here is that we'll at least try to optimize for when
-                # we gone below the visible clipping area.
-                #
-                # [[[TODO: WDW - would be nice to optimize this better.
-                # for example, perhaps we can assume the caret will always
-                # be visible, and we can start our text search from there.
-                # Logged as bugzilla bug 319771.]]]
-                #
-                break
+        for string, startOffset, endOffset in lines:
+            zones.extend(self.splitTextIntoZones(accessible, string, startOffset, cliprect))
 
         return zones
 
@@ -738,21 +663,21 @@ class Context:
         # right or nuke it.
 
         indicatorExtents = [extents.x, extents.y, 1, extents.height]
-        role = accessible.getRole()
-        if role == pyatspi.ROLE_TOGGLE_BUTTON:
+        role = AXObject.get_role(accessible)
+        if role == Atspi.Role.TOGGLE_BUTTON:
             zone = StateZone(accessible, *indicatorExtents, role=role)
             if zone:
                 zones.insert(0, zone)
             return
 
-        if role == pyatspi.ROLE_TABLE_CELL \
+        if role == Atspi.Role.TABLE_CELL \
            and self.script.utilities.hasMeaningfulToggleAction(accessible):
-            role = pyatspi.ROLE_CHECK_BOX
+            role = Atspi.Role.CHECK_BOX
 
-        if role not in [pyatspi.ROLE_CHECK_BOX,
-                        pyatspi.ROLE_CHECK_MENU_ITEM,
-                        pyatspi.ROLE_RADIO_BUTTON,
-                        pyatspi.ROLE_RADIO_MENU_ITEM]:
+        if role not in [Atspi.Role.CHECK_BOX,
+                        Atspi.Role.CHECK_MENU_ITEM,
+                        Atspi.Role.RADIO_BUTTON,
+                        Atspi.Role.RADIO_MENU_ITEM]:
             return
 
         zone = None
@@ -781,23 +706,27 @@ class Context:
 
         try:
             component = accessible.queryComponent()
-            extents = component.getExtents(pyatspi.DESKTOP_COORDS)
-        except:
+            extents = component.getExtents(Atspi.CoordType.SCREEN)
+        except Exception:
             return []
 
         try:
-            role = accessible.getRole()
-        except:
+            role = AXObject.get_role(accessible)
+        except Exception:
             return []
 
         zones = self.getZonesFromText(accessible, cliprect)
-        if not zones and role in [pyatspi.ROLE_SCROLL_BAR,
-                                  pyatspi.ROLE_SLIDER,
-                                  pyatspi.ROLE_PROGRESS_BAR]:
+        if not zones and role in [Atspi.Role.SCROLL_BAR,
+                                  Atspi.Role.SLIDER,
+                                  Atspi.Role.PROGRESS_BAR]:
             zones.append(ValueZone(accessible, *extents))
         elif not zones:
-            string = self.script.speechGenerator.getName(accessible)
-            useless = [pyatspi.ROLE_TABLE_CELL, pyatspi.ROLE_LABEL]
+            string = ""
+            redundant = [Atspi.Role.TABLE_ROW]
+            if role not in redundant:
+                string = self.script.speechGenerator.getName(accessible, inFlatReview=True)
+
+            useless = [Atspi.Role.TABLE_CELL, Atspi.Role.LABEL]
             if not string and role not in useless:
                 string = self.script.speechGenerator.getRoleName(accessible)
             if string:
@@ -814,7 +743,64 @@ class Context:
         if child == parent:
             return True
 
-        return pyatspi.findAncestor(child, lambda x: x == parent)
+        return AXObject.find_ancestor(child, lambda x: x == parent)
+
+    def setCurrentToZoneWithObject(self, obj):
+        """Attempts to set the current zone to obj, if obj is in the current context."""
+
+        tokens = ["FLAT REVIEW: Current", self.getCurrentAccessible(),
+                  f"line: {self.lineIndex}, zone: {self.zoneIndex},",
+                  f"word: {self.wordIndex}, char: {self.charIndex})"]
+        debug.printTokens(debug.LEVEL_INFO, tokens, True)
+
+        zone = self._findZoneWithObject(obj)
+        tokens = ["FLAT REVIEW: Zone with", obj, "is", zone]
+        debug.printTokens(debug.LEVEL_INFO, tokens, True)
+        if zone is None:
+            return False
+
+        for i, line in enumerate(self.lines):
+            if zone in line.zones:
+                self.lineIndex = i
+                self.zoneIndex = line.zones.index(zone)
+                word, offset = zone.wordWithCaret()
+                if word:
+                    self.wordIndex = word.index
+                    self.charIndex = offset
+                msg = "FLAT REVIEW: Updated current zone."
+                debug.printMessage(debug.LEVEL_INFO, msg, True)
+                break
+        else:
+            msg = "FLAT REVIEW: Failed to update current zone."
+            debug.printMessage(debug.LEVEL_INFO, msg, True)
+            return False
+
+        tokens = ["FLAT REVIEW: Updated", self.getCurrentAccessible(),
+                  f"line: {self.lineIndex}, zone: {self.zoneIndex},",
+                  f"word: {self.wordIndex}, char: {self.charIndex})"]
+        debug.printTokens(debug.LEVEL_INFO, tokens, True)
+        return True
+
+    def _findZoneWithObject(self, obj):
+        """Returns the existing zone which contains obj."""
+
+        if obj is None:
+            return None
+
+        for zone in self.zones:
+            if zone.accessible == obj:
+                return zone
+
+            # Some items get pruned from the flat review tree. For instance, a
+            # tree item which has a descendant section whose text is the displayed
+            # text of the tree item, that section will be in the flat review tree
+            # but the ancestor item might not.
+            if AXObject.is_ancestor(zone.accessible, obj):
+                tokens = ["FLAT REVIEW:", zone.accessible, "is ancestor of zone accessible", obj]
+                debug.printTokens(debug.LEVEL_INFO, tokens, True)
+                return zone
+
+        return None
 
     def getShowingZones(self, root, boundingbox=None):
         """Returns an unsorted list of all the zones under root and the focusZone."""
@@ -823,8 +809,8 @@ class Context:
             boundingbox = self.bounds
 
         objs = self.script.utilities.getOnScreenObjects(root, boundingbox)
-        msg = "FLAT REVIEW: %i on-screen objects found for %s" % (len(objs), root)
-        debug.println(debug.LEVEL_INFO, msg, True)
+        tokens = ["FLAT REVIEW:", len(objs), "on-screen objects found for", root]
+        debug.printTokens(debug.LEVEL_INFO, tokens, True)
 
         allZones, focusZone = [], None
         for o in objs:
@@ -842,8 +828,8 @@ class Context:
                 zones = list(filter(lambda z: z.hasCaret(), zones)) or zones
                 focusZone = zones[0]
 
-        msg = "FLAT REVIEW: %i zones found for %s" % (len(allZones), root)
-        debug.println(debug.LEVEL_INFO, msg, True)
+        tokens = ["FLAT REVIEW:", len(allZones), "zones found for", root]
+        debug.printTokens(debug.LEVEL_INFO, tokens, True)
         return allZones, focusZone
 
     def clusterZonesByLine(self, zones):
@@ -872,8 +858,8 @@ class Context:
                 zone.line = lines[lineIndex]
                 zone.index = zoneIndex
 
-        msg = "FLAT REVIEW: Zones clustered into %i lines" % len(lines)
-        debug.println(debug.LEVEL_INFO, msg, True)
+        tokens = ["FLAT REVIEW: Zones clustered into", len(lines), "lines"]
+        debug.printTokens(debug.LEVEL_INFO, tokens, True)
         return lines
 
     def getCurrent(self, flatReviewType=ZONE):
@@ -889,12 +875,12 @@ class Context:
         current = zone
         if flatReviewType == Context.LINE:
             current = zone.line
-        elif zone.words:
+        elif flatReviewType != Context.ZONE and zone.words:
             current = zone.words[self.wordIndex]
             if flatReviewType == Context.CHAR and current.chars:
                 try:
                     current = current.chars[self.charIndex]
-                except:
+                except Exception:
                     return None, -1, -1, -1, -1
 
         return current.string, current.x, current.y, current.width, current.height
@@ -917,7 +903,7 @@ class Context:
 
     def _getClickPoint(self):
         string, x, y, width, height = self.getCurrent(Context.CHAR)
-        if x < 0 and y < 0:
+        if (x < 0 and y < 0) or (width <= 0 and height <=0):
             return -1, -1
 
         # Click left of center to position the caret there.
@@ -933,18 +919,19 @@ class Context:
         if x < 0 or y < 0:
             return False
 
-        eventsynthesizer.routeToPoint(x, y)
-        return True
+        return AXEventSynthesizer.route_to_point(x, y)
 
     def clickCurrent(self, button=1):
         """Performs a mouse click on the current accessible."""
 
         x, y = self._getClickPoint()
-        if x < 0 or y < 0:
-            return False
+        if x >= 0 and y >= 0 and AXEventSynthesizer.click_point(x, y, button):
+            return True
 
-        eventsynthesizer.clickPoint(x, y, button)
-        return True
+        if AXEventSynthesizer.click_object(self.getCurrentAccessible(), button):
+            return True
+
+        return False
 
     def _getCurrentZone(self):
         if not (self.lines and 0 <= self.lineIndex < len(self.lines)):
@@ -987,6 +974,7 @@ class Context:
                 regionWithFocus = zone.brailleRegion
                 regionWithFocus.cursorOffset = 0
                 if zone.words:
+                    regionWithFocus.cursorOffset += zone.words[0].startOffset - zone.startOffset
                     for wordIndex in range(0, self.wordIndex):
                         regionWithFocus.cursorOffset += \
                             len(zone.words[wordIndex].string)
@@ -1097,7 +1085,7 @@ class Context:
         """
 
         if not self.lines:
-            debug.println(debug.LEVEL_FINE, 'goPrevious(): no lines in context')
+            debug.printMessage(debug.LEVEL_INFO, 'goPrevious(): no lines in context')
             return False
 
         moved = False
@@ -1230,7 +1218,7 @@ class Context:
         """
 
         if not self.lines:
-            debug.println(debug.LEVEL_FINE, 'goNext(): no lines in context')
+            debug.printMessage(debug.LEVEL_INFO, 'goNext(): no lines in context')
             return False
 
         moved = False
@@ -1248,12 +1236,14 @@ class Context:
                     self.wordIndex = 0
                     self.charIndex = 0
                     moved = True
+                    braille.clear()
                 elif wrap & Context.WRAP_TOP_BOTTOM:
                     self.lineIndex  = 0
                     self.zoneIndex  = 0
                     self.wordIndex = 0
                     self.charIndex = 0
                     moved = True
+                    braille.clear()
         elif flatReviewType == Context.CHAR:
             zone = self.lines[self.lineIndex].zones[self.zoneIndex]
             if zone.words:
